"""
Version information.
"""

# the major version
pyosmium_major = '2.12'
# current release (Pip version)
pyosmium_release = '2.12.0'

# libosmium version shipped with the Pip release
libosmium_version = '2.12.0'
